/* Original code and idea (C) Antonio Jos Villena Godoy, updates and modifications (C) Carlos Snchez */

/*
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/


/* This class contains the SNA to WAV converter */

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.RandomAccessFile;






public class converter {


	
	// Commandline orders
	commandLine clOrder;

	// The input and output files
	DataInputStream inputFile;

	RandomAccessFile outputFile;

	static boolean bit;

	/*
	 * The Ultraload Code, Z80 code for the loader
	 */
	static byte ultraCode[] = StringToByteArray(Z80Code.ultraCode);

	/*
	 * The basic loader code
	 */
	static byte basicCode[] = StringToByteArray("\23\0\0\0          \170\0\0\0\16\0 \172\0\377\0\0\12\0\371\300\60\16\0\0\331\134\0\15");

	/*
	 * Total Block that contains both ultraloader code and basic code
	 */
	static byte totalBlock[] = new byte[basicCode.length + ultraCode.length - 3];

	/*
	 * Z80 Code that restores registry status once the ultraload finishes, it is
	 * usually stored at printer's buffer, but can be stored on screen if -s
	 * option is issued
	 */
	static byte snaCode[] = StringToByteArray(Z80Code.snaCode);

	// Array to store the SNA file memory information (48k)
	byte RAM[] = new byte[49152];
	

	private void saveRAM() throws IOException
	{
		for (int i = 0; i < RAM.length; i++)
			for (int j = 0; j < 8; j++, RAM[i] <<= 1)
				pulse((RAM[i] & 0x80) == 0 ? 2 : 4);

	}
	
	 /* Generates standar load . It is called twice, once for each block. Data is
	  * obtained from TotalBlock array. 
	 */
	private void datnor(int i, int j) throws IOException {
		pulses(2, 6);
		while (i++ < j)
			for (int k = 0; k < 8; k++, totalBlock[i] <<= 1)
				pulses(2, (totalBlock[i] & 0x80) == 0 ? clOrder.LOBitValue : clOrder.HIBitValue);
	}

	/*
	 * Writes into output file "sampleCount" samples. Every execution it changes 
	 * the value of "bit". This is the only function that actually stores saves
	 * to file
	 */
	private void pulse(int sampleCount) throws IOException {
		if (bit) {
			if (clOrder.isMono)
				while (sampleCount-- > 0)
					outputFile.writeByte(0x10);
			else
				while (sampleCount-- > 0)
					outputFile.writeInt(0x108A108A);
		} else {
			if (clOrder.isMono)
				while (sampleCount-- > 0)
					outputFile.writeByte(0xF0);
			else
				while (sampleCount-- > 0)
					outputFile.writeInt(0xF075F075);
		}
		bit ^= true;
	}
	
	
	/*
	 * Generates "pulseCount" samples of "sampleCount" type
	 */
	private void pulses(int pulseCount, int sampleCount) throws IOException {
		while (pulseCount-- > 0)
			pulse(sampleCount);
	}

	
	
	/*
	 * Obtains a byte array from a string
	 */
	private static byte[] StringToByteArray(String str) {
		byte result[] = new byte[str.length()];
		for (int i = 0; i < str.length(); i++)
			result[i] = (byte) str.charAt(i);
		return result;
	}

	/*
	 * Copies some data from one array to another one.
	 */
	private void move(byte[] srcArray, byte[] destArray, int srcIndex,
			int destIndex, int length) {
		for (int i = 0; i < length; i++)
			destArray[i + destIndex] = srcArray[i + srcIndex];
	}

	/*
	 * Reads a word (16 bits) from the input file and stores it at position n on
	 * the snaCode
	 */
	private void readWord(int n) throws IOException {
		snaCode[n++] = inputFile.readByte();
		snaCode[n] = inputFile.readByte();
	}


public void convert(commandLine clOrder, File f) {

	this.clOrder = clOrder;

	try {

		System.out.print("Processing file: " + f.getName() + "... ");
		inputFile = new DataInputStream(new FileInputStream(f));

		if (inputFile.available() != 49179) 
		{
			System.out.println("Error: The .SNA file is invalid or corrupt");
			System.exit(1);
		} 
		else
		{
			
			outputFile = new RandomAccessFile(f.getAbsolutePath().substring(0,f.getAbsolutePath().indexOf('.')).toLowerCase() + ".wav", "rw");

			// The value that modifies everything on the ultraload to
			// make them Timex TC2048 compatible
			int tc2048Delta = clOrder.isTC2048 ? 0xA0 : 0;

			// Determine where to put sna regsitry restore code
			// (1B00 = printer's buffer, 1751 = at the end of the screen)
			int snaCodeLocation = clOrder.useScreen ? 0x1751 : 0x1B00;

			// change the ultraCode and snaCode code so it jumps to the
			// correct address
			ultraCode[85] = (byte) (snaCodeLocation & 0xFF);
			snaCode[29] = ultraCode[86] = (byte) (0x40 | snaCodeLocation >> 8);
			snaCode[28] = (byte) (0x55 + snaCodeLocation & 0xFF);

			// modifies ultracode and snacode to make them Sinclair/Timex
			// compatible
			ultraCode[24] = snaCode[31] = (byte) (0xBF - tc2048Delta);
			ultraCode[9] = snaCode[39] = (byte) (0xF8 - tc2048Delta);
			snaCode[45] = (byte) (0xB8 - tc2048Delta);
			ultraCode[108] = snaCode[52] = (byte) (0xFF - tc2048Delta);
			ultraCode[17] = (byte) (0xD6 - tc2048Delta);
			ultraCode[27] = (byte) (0xD4 - tc2048Delta);
			ultraCode[30] = (byte) (0xD0 - tc2048Delta);
			ultraCode[34] = (byte) (0x03 + tc2048Delta * 96 / 160);
			ultraCode[35] = (byte) (0x82 - tc2048Delta / 160);
			ultraCode[46] = (byte) (0xC3 - tc2048Delta);
			ultraCode[72] = (byte) (0xC0 - tc2048Delta);

			// Read I registry value and stores it in snaCode
			snaCode[4] = inputFile.readByte();

			readWord(8); // same for HL'
			readWord(11);// same for DE'
			readWord(24);// same for BC'
			readWord(14);// same for AF'
			readWord(58);// same for HL
			readWord(61);// same for DE
			readWord(64);// same for BC
			readWord(68);// same for IY
			readWord(72);// same for IX

			// Reads Interrupt
			if (inputFile.readUnsignedByte() < 128)
				snaCode[74] = (byte) 0xF3;

			// Reads R register, substracts 5 for it to have the same value (5 =>  LD R,A, PUSH PC, RET) 
			snaCode[78] = (byte) (inputFile.readByte() - 5);

			readWord(20); // same for AF

			// Obtains SP
			int SP = (inputFile.readUnsignedByte() | (inputFile
					.readUnsignedByte() << 8)) + 2;

			// And store it into the snapCode. readWord is not used
			// as we need to store also the value at SP var
			snaCode[1] = (byte) (SP & 0xFF);
			snaCode[2] = (byte) (SP >> 8);

			// Reads interruption mode
			int interrupt = inputFile.readByte();
			// changes the sna code to take care of it
			if (interrupt != 0)
				snaCode[76] += (interrupt + 1) << 3;

			// reads and ignores border color
			inputFile.readByte();

			// Load RAM from SNA file
			inputFile.read(RAM, 0, 49152);

			// Saves RAM that is going to be overwritten by SNA Code
			move(RAM, snaCode, SP - 0x4002, 83, 2);
			move(RAM, RAM, 0x40BF - tc2048Delta, snaCode.length
					+ snaCodeLocation, 4);
			move(RAM, RAM, 0x40F8 - tc2048Delta, snaCode.length
					+ snaCodeLocation + 4, 30);
			move(RAM, RAM, 0x41B8 - tc2048Delta, snaCode.length
					+ snaCodeLocation + 34, 30);
			move(RAM, RAM, 0x41FF - tc2048Delta, snaCode.length
					+ snaCodeLocation + 64, 4);

			// Moves ultracode to RAM
			move(ultraCode, RAM, 0, 0x40BF - tc2048Delta, 4);
			move(ultraCode, RAM, 80, 0x40F8 - tc2048Delta, 30);
			move(ultraCode, RAM, 80, 0x41B8 - tc2048Delta, 30);
			move(ultraCode, RAM, 0, 0x41FF - tc2048Delta, 4);

			// Some changes for Timex
			RAM[0x41D4 - tc2048Delta] = (byte) (0xBF - tc2048Delta);
			RAM[0x41D5 - tc2048Delta]--;
			RAM[0x41D0 - tc2048Delta]++;

			// Copies SNA registry restore code to RAM
			move(snaCode, RAM, 0, snaCodeLocation, snaCode.length);
			move(basicCode, totalBlock, 0, 0, basicCode.length);
			move(ultraCode, totalBlock, 4, basicCode.length,
					ultraCode.length - 4);

			// Obtain Snapshot name, up to the dot
			String snapName = f.getName().substring(0,
					f.getName().lastIndexOf('.'));
			// Ensure name is exactly 10 characters length
			if (snapName.length() > 10)
				snapName = snapName.substring(1, 10);
			while (snapName.length() < 10)
				snapName += " ";

			// Fill up the header with the snapshot name
			move(StringToByteArray(snapName), totalBlock, 0, 4, 10);
			int i = 0;
			for (int k = 0; k < 2; k++) {
				int c = 0;
				int j = totalBlock[i++] | totalBlock[i++] << 8;
				while (--j > 0)
					c = c ^ totalBlock[i++];
				totalBlock[i++] = (byte) c;
			}

			// Start writing the WAV file

			// WAV Header
			if (clOrder.isMono)
				outputFile
						.writeBytes("RIFF    WAVEfmt \20\0\0\0\1\0\1\0\104\254\0\0\104\254\0\0\1\0\10\0data    ");
			else
				outputFile
						.writeBytes("RIFF    WAVEfmt \20\0\0\0\1\0\2\0\104\254\0\0\020\261\2\0\4\0\20\0data    ");

			// Start Pulse
			pulses(clOrder.leadToneLength, 23);
			datnor(1, 20);

			// Header
			pulses(clOrder.leadToneLength, 23);
			datnor(22, totalBlock.length - 1);

			pulses(6, 40);
			pulse(2);
			pulse(4);

			// actual SNA file data (full speccy's memory)
			saveRAM();

			//pulses(2, 2);
			pulses(11025, 4);
			pulses(1,44100);
			outputFile.setLength(outputFile.getFilePointer());
			i = (int) outputFile.length();
			outputFile.seek(4);
			outputFile.writeInt(Integer.reverseBytes(i - 8));
			outputFile.seek(40);
			outputFile.writeInt(Integer.reverseBytes(i - 44));
			outputFile.close();
			System.out.println("Succesfully converted");
		}
		inputFile.close();
	} catch (IOException e2) {
		System.out.println("Can't save file. File already in use?");
	}
}
}
